package turtleWorld;

import grobs.* ;
import java.awt.* ;
import java.util.*;

public class HanoiController extends Controller  {
	private final int WIDTHSTEP = 20 ;
	private final int HEIGHT = 20 ;
	private final int GAP = 5 ;
	private final int[] SPINDLE_X_MIN = {100, 300, 500 } ;
	private final int[] SPINDLE_X_MAX = {290, 490, 690 } ;
	private final int SPINDLE_Y_MIN = 50 ;
	private final int SPINDLE_Y_MAX = 300 ;
	private boolean locked = false ; // locked is used to make sure only one UI action is processed at a time.

	private Color colours[] = new Color[]{ Color.BLUE, Color.CYAN, Color.GREEN,
			Color.YELLOW, Color.ORANGE, Color.RED, Color.MAGENTA } ;
	Stack disks = new Stack() ;
	ShapeGrob[] spindleRects = new ShapeGrob[3] ;
	ShapeGrob[] spindleImage = new ShapeGrob[3] ;
	Stack[] stacks = new Stack[3] ;

	public HanoiController( GrobArena arena, Log log, ThreadCounter threadCounter ) {
		super( arena, log, threadCounter ) ;
		Color backgroundColor = Color.WHITE ;
		arena.setBackground( backgroundColor );
		
		// Make a Base
		int baseThickness = 20 ;
		Rectangle baseRect = new Rectangle( SPINDLE_X_MIN[0], SPINDLE_Y_MAX - baseThickness,
				SPINDLE_X_MAX[2]-SPINDLE_X_MIN[0], baseThickness  ) ;
		ShapeGrob base = new ShapeGrob( baseRect ) ;
		base.setZIndex(-1) ;
		base.setDraggable(false) ;
		base.setDropTarget(false) ;
		base.setFillColor( Color.GRAY ) ;
		base.setOutlineColor( Color.GRAY ) ; 
		arena.add( base ) ;
		
		for( int i=0 ; i < 3 ; ++i ) {
			// The area around a spindle serves as a drop target. This is invisible, being the 
			// Sane colour as the back ground.
			Rectangle r = new Rectangle( SPINDLE_X_MIN[i], SPINDLE_Y_MIN,
					SPINDLE_X_MAX[i]-SPINDLE_X_MIN[i], SPINDLE_Y_MAX-SPINDLE_Y_MIN  ) ;
			spindleRects[i] = new ShapeGrob( r ) ;
			spindleRects[i].setZIndex(-2) ;
			spindleRects[i].setDraggable(false) ;
			spindleRects[i].setDropTarget(true) ;
			spindleRects[i].setFillColor( backgroundColor ) ;
			spindleRects[i].setOutlineColor( backgroundColor ) ; 
			arena.add( spindleRects[i] ) ;
			
			// What the user sees is a just a spike
			int middle = ( SPINDLE_X_MIN[i] + SPINDLE_X_MAX[i]) / 2 ;
			int halfWidth = 3 ;
			r = new Rectangle( middle - halfWidth, SPINDLE_Y_MIN,
					2 * halfWidth, SPINDLE_Y_MAX-SPINDLE_Y_MIN  ) ;
			spindleImage[i] = new ShapeGrob( r ) ;
			spindleImage[i].setZIndex(-1) ;
			spindleImage[i].setDraggable(false) ;
			spindleImage[i].setDropTarget(false) ;
			spindleImage[i].setFillColor(Color.GRAY) ;  //xyz
			spindleImage[i].setOutlineColor(Color.GRAY) ; 
			arena.add( spindleImage[i] ) ;
			
			// Make one stack.
			stacks[i] = new Stack() ; }
		for(int i = 0 ; i < 7 ; ++i ) addDisk() ;
		log.println( "Can you move all the disks to the rightmost spindle?") ;
		log.println( "*  One one disk at a time can be moved.") ;
		log.println( "*  Only the top disk on each stack can be moved.") ;
		log.println( "*  Larger disks cannot go on smaller.") ;
	}

	@Override
	public String controllerName() {
		return "Towers of Hanoi" ;
	}

	@Override
	public String[] buttons() {
		return new String[] {
				"reset", "scramble", "more", "fewer", "solve" } ; }

	@Override
	public String[] checkBoxes() {
		return new String[] {"cheat"} ; }

	@Override
	public String[] startUp() {
		return new String[] {  "reset" } ; }

	void addDisk() {
		int i = disks.size() ;
		new Disk(i,0, arena, disks) ;
	}

	public void solve() {
		if( locked ) return ;
		locked = true ;
		solve(2, disks.size() ) ;
		locked = false ;
	}

	/** Move the n smallest disks to spindle dest
	 * 
	 * @param dest
	 * @param n
	 */
    void solve(int dest, int n) {
		// TODO
    }



	private void move(Disk d, int dest) {
		d.moveTo(dest) ;
		double t = 0.2 ;
		d.startFlight(t) ; pause(t) ;
		d.endFlight() ;
	}

	public void more() {
		if( locked ) return ;
		locked = true ;
		addDisk() ;
		for( Disk d : stacks[0] ) d.startFlight(1) ;
		pause(1) ;
		for( Disk d : stacks[0] ) d.endFlight() ;
		locked = false ;
	}

	public void fewer() {
		if( locked ) return ;
		locked = true ;
		if( disks.size() == 0 ) return ;
		Disk disk = disks.get(disks.size() - 1) ;
		Stack stack = stacks[ disk.spindleNum ] ;
		disk.retire() ;
		int sz = stack.size() ;
		for( int i=sz-1 ; i >= 0 ; --i ) { stack.get(i).startFlight(1) ; pause(0.1) ;} 
		pause(1.0 - sz*0.1) ;
		for( int i=sz-1 ; i >= 0 ; --i ) { stack.get(i).endFlight() ; pause(0.1) ; }
		locked = false ;
	}

	public void scramble() {
		if( locked ) return ;
		locked = true ;
		Random rng = new Random();
		for( Disk d : disks ) {
			int s = rng.nextInt(3) ;
			d.moveTo(s) ; }
		for( Disk d : disks) d.startFlight(1) ;
		pause(1) ;
		for( Disk d : disks) d.endFlight() ;
		locked = false ;
	}

	public void reset() {
		if( locked ) return ;
		locked = true ;
		for( Disk d : disks ) d.moveTo(0) ; 
		for( Disk d : disks) d.startFlight(1) ;
		pause(1) ;
		for( Disk d : disks) d.endFlight() ;
		locked = false ;
	}
	
	boolean cheatMode = false ;
	
	public void cheatOn() {
		cheatMode = true ;
	}
	
	public void cheatOff() {
		cheatMode = false ;
	}

	public void onEndDrag( Grob dragged, Grob dropTarget ) {
		if( locked ) return ;
		locked = true ;
		Disk disk = (Disk) dragged ;
		int spindle = -1 ;
		for( int s = 0; s < 3 && spindle == -1 ; ++s ) {
			if( spindleRects[s] == dropTarget ) {
				spindle = s ; } }
		ArrayList<Disk> flyingDisks = new ArrayList<Disk>() ;
		flyingDisks.add( disk ) ;
		int sourceSpindle =  disk.spindleNum ;
		Stack sourceStack = stacks[ sourceSpindle ] ;
		if( spindle == -1 || spindle == disk.spindleNum ) {
			// disk.moveTo( disk.spindleNum ) ; // Back to where it came from
		} else {
			Stack stack = stacks[ spindle ] ;
			if( !cheatMode && sourceStack.get(0) != disk ) {
				log.println( "Only the disk on the top of the stack can be moved") ;
			} else if( stack.isEmpty() || stack.get(0).diskNum > disk.diskNum ) {
				if( sourceStack.get(0) != disk) {
					log.println( "I really shouldn't let you do this!" ) ;
					// All the disks on the source stack that are smaller than
					// the dragged disk will be moved to.
					for( Disk d : sourceStack ) {
						if( d.diskNum == disk.diskNum ) break ;
						flyingDisks.add( d ) ;
					}
				}
				// Now we actually move the dragged disk and also and that are coming along because of
				// cheat mode.
				for( Disk d : flyingDisks ) {
					d.moveTo( spindle ) ;
				}
			} else {
				log.println( "Larger disks cannot go on smaller") ;
				// disk.moveTo(disk.spindleNum) ;
			} }
		for( Disk d : flyingDisks ) { d.startFlight(1) ; } 
		pause(1) ;
		for( Disk d : flyingDisks ) { d.endFlight() ; }
		locked = false ;
	}

	class Disk extends ShapeGrob {
		final int diskNum ; // 0 for smallest
		private int spindleNum  ;

		Disk(int i, int s, GrobArena arena, Stack disks) {
			super( new Rectangle(WIDTHSTEP*(i+1), 20) ) ;
			setZIndex(i) ;
			setFillColor( colours[i % colours.length] ) ;
			diskNum = i ;
			addTo(s) ;
			setDraggable( true ) ; //xyz 
			disks.add(this) ;
			arena.add(this) ;
		}

		void retire() {
			Stack stack = stacks[spindleNum] ;
			stack.remove(this) ;
			disks.remove(this);
			arena.remove(this) ;
			//xyz for( Disk d : stack) d.setDraggable(false) ;
			//xyz if( ! stack.isEmpty() ) stack.get(0).setDraggable(true) ;
		}

		private void addTo(int s) {
			spindleNum = s ;
			Stack stack =  stacks[s] ;
			stack.add(this) ;
			Collections.sort(stack, new Comparator<Disk>() {
				@Override public int compare(Disk o1, Disk o2) {
					return o1.diskNum - o2.diskNum ; } } ) ;
			// Only the top can be draggable
			//xyz for( Disk d : stack) d.setDraggable(false) ;
			//xyz if( ! stack.isEmpty() ) stack.get(0).setDraggable(true) ;
		}

		void moveTo( int s ) {
			if( s != spindleNum ) {
				Stack stack =  stacks[spindleNum] ;
				stack.remove(this) ;
				//xyz for( Disk d : stack) d.setDraggable(false) ;
				//xyz if( ! stack.isEmpty() ) stack.get(0).setDraggable(true) ;
				addTo(s) ;
			}
		}

		double myX() {
			int middle = (SPINDLE_X_MIN[spindleNum] + SPINDLE_X_MAX[spindleNum])/2 ;
			return  middle - (diskNum+1)*WIDTHSTEP/2 ;
		}

		double myY() {
			Stack stack = stacks[spindleNum] ;
			Assert.assertTrue( stack.indexOf(this) != -1, "spindleNum is wrong for disk" ) ;
			int height = stack.size() - stack.indexOf(this) ;
			return SPINDLE_Y_MAX - (height+1)*(GAP+HEIGHT) ;
		}

		void startFlight(double fightTime) {
			double x = myX() ;
			double y = myY() ;
			double x0 = this.getPositionX() ;
			double y0 = this.getPositionY() ;
			this.setVelocity((x-x0)/fightTime, (y-y0)/fightTime) ;
		}

		void endFlight() {
			this.setVelocity(0, 0) ;
			this.setPosition(myX(), myY()) ;
		}
	}

	class Stack extends ArrayList<Disk> {} 
}